import os
from earth_osm.eo import get_osm_data
from earth_osm.utils import tags_melt, tags_explode, columns_melt

def test_column_melt(shared_data_dir):
    region = "nigeria"
    primary_name = "power"
    feature_name = "substation"
    mp = True
    update = False
    data_dir = shared_data_dir

    df = get_osm_data(
        region, 
        primary_name, 
        feature_name,
        data_dir=data_dir
        )

    df_cols = list(df.columns)

    df_col_melt = columns_melt(df.copy(), ['refs'])
    assert df_cols == list(df.columns) # check if original df is same
    df_col_melt_cols = list(df_col_melt.columns)

    df_exp  = tags_explode(df_col_melt.copy())
    assert df_col_melt_cols == list(df_col_melt.columns) # check if original df is same
    # print(df_exp.columns)

    # print difference in columns
    print(set(df_exp.columns.tolist()) - set(df.columns.tolist()))
    print(set(df.columns.tolist()) - set(df_exp.columns.tolist()))

    assert set(df_exp.columns.tolist()) == set(df.columns.tolist())

    # check the types of the column refs, row by row
    for i in range(len(df)):
        if not df['refs'].iloc[i] == df_exp['refs'].iloc[i]:
            # check if nan
            if str(df['refs'].iloc[i]) == 'nan' and str(df_exp['refs'].iloc[i]) == 'nan':
                continue
            
            print(f'Row {i} is not equal')
            i_ = i
            print(df['refs'].iloc[i_])
            print(df_exp['refs'].iloc[i_])

            print(type(df['refs'].iloc[i_]))
            print(type(df_exp['refs'].iloc[i_]))


    for col in df.columns:
        assert df[col].equals(df_exp[col]), f'Column {col} is not equal'

    
def test_melt(shared_data_dir):
    region = "nigeria"
    primary_name = "power"
    feature_name = "substation"
    mp = True
    update = False
    data_dir = shared_data_dir

    df = get_osm_data(
        region, 
        primary_name, 
        feature_name,
        data_dir=data_dir
        )
    
    df_cols = list(df.columns)

    # melt tags
    df_melt = tags_melt(df.copy(), nan_threshold=0.75)
    assert df_cols == list(df.columns) # check if original df is same
    df_melt_cols = list(df_melt.columns)
    assert 'other_tags' in df_melt.columns
    # print(df_melt['other_tags'])

    # explode tags
    df_explode = tags_explode(df_melt.copy())
    assert df_melt_cols == list(df_melt.columns) # check if original df is same
    # print(df_explode.columns)

    # print difference
    print(set(df.columns.tolist()) - set(df_explode.columns.tolist()))
    print(set(df_explode.columns.tolist()) - set(df.columns.tolist()))

  
    assert set(df.columns.tolist()) == set(df_explode.columns.tolist())    

    # print length of columns for all
    print(f'Length of df columns: {len(df.columns)}')
    print(f'Length of df_melt columns: {len(df_melt.columns)}')
    print(f'Length of df_explode columns: {len(df_explode.columns)}')

    # check if each column has the same length and values, df_exp and df should be same
    for col in df.columns:
        assert df[col].equals(df_explode[col])



if __name__ == '__main__':
    test_column_melt()
    test_melt()


from earth_osm.eo import save_osm_data
import os


def test_building():
    save_osm_data(
        region_list=["malta"],
        primary_name="building",
        feature_list=['ALL'],
        update=False,
        mp=True,
        data_dir=os.path.join(os.getcwd(), "earth_data_test"),
        out_dir=os.path.join(os.getcwd(), "earth_data_test"),
        out_format=["csv", "geojson"],
        out_aggregate=False,
    )


if __name__ == '__main__':
    test_building()



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""

This module provides functions to fetch and process data from the OpenStreetMap Taginfo API.
The resulting file is stored in the internal data directory as `earth_data.json`.

"""

import json
import os
import pandas as pd
import requests

WIKI_PRIMARY_LIST = [
    'aerialway',
    'aeroway',
    'amenity',
    'barrier',
    'boundary',
    'building',
    'craft',
    'emergency',
    'geological',
    'highway',
    'historic',
    # 'landuse', # high in relations
    'leisure',
    'man_made',
    'military',
    # 'natural',
    'office',
    'place',
    'power',
    'public_transport',
    'railway',
    # 'route', # high in relations
    'shop',
    'sport',
    'tourism',
    # 'water',
    'waterway'
]

BASE_URL = 'https://taginfo.openstreetmap.org/api/4'


def fetch_data_from_api(path: str, params: dict) -> dict:
    url = BASE_URL + path
    response = requests.get(url, params=params)
    response.raise_for_status()
    return response.json()


def get_key_overview(key: str) -> dict:
    key_data = fetch_data_from_api('/key/overview', {'key': key})['data']
    return {key: {'en_description': key_data['description']['en']['text']}}


def get_wiki_features_df(key: str) -> pd.DataFrame:
    tags_list = fetch_data_from_api('/tags/list', {'key': key})['data']

    # Convert to DataFrame for easier manipulation
    tags_df = pd.DataFrame(tags_list)

    # Sort values
    tags_df.sort_values(by=['count_all'], ascending=False, inplace=True)

    # Add en_description column at position 2 and drop None
    tags_df['en_description'] = tags_df.apply(lambda row: row['wiki']['en']['description'] if 'wiki' in row and 'en' in row['wiki'] and 'description' in row['wiki']['en'] else None, axis=1)
    tags_df = tags_df[~tags_df['en_description'].isna()] 
    tags_df.insert(2, 'en_description', tags_df.pop('en_description'))

    # Drop unwanted columns
    drop_columns = ['in_wiki', 'wiki', 'count_all_fraction', 'count_nodes_fraction', 'count_ways_fraction', 'count_relations_fraction']
    tags_df.drop(columns=drop_columns, inplace=True, errors='ignore')

    return tags_df

def get_wiki_features(key: str) -> dict:
    tags_df = get_wiki_features_df(key)

    # Convert DataFrame to list of dictionaries
    tags_list = tags_df.to_dict(orient='records')
    
    # Convert list of dictionaries to nested dictionary
    tags_dict = {tag["value"]: tag for tag in tags_list}
    
    return {key: tags_dict}


def fetch_all_data():
    primary_keys_data = {key: get_key_overview(key)[key] for key in WIKI_PRIMARY_LIST}
    features_data = {key: get_wiki_features(key)[key] for key in WIKI_PRIMARY_LIST}

    # Combining the dictionaries
    all_data = primary_keys_data.copy()
    for key in WIKI_PRIMARY_LIST:
        all_data[key]['features'] = features_data[key]

    return primary_keys_data

def save_data(all_data: dict, save_path: str):
    with open(save_path, 'w') as fp:
        json.dump(all_data, fp, indent=4)

def get_data(data_filepath: str, update: bool = False) -> dict:
    if not os.path.exists(data_filepath) or update:
        save_data(fetch_all_data(), data_filepath)

    with open(data_filepath, 'r') as fp:
        return json.load(fp)

def get_tag_data():
    data_dir = os.path.join(os.path.dirname(__file__), 'data')
    json_filepath = os.path.join(data_dir, 'earth_data.json')

    if not os.path.exists(json_filepath):
        raise FileNotFoundError(f'File {json_filepath} not found. Please run `python -m earth_osm.taginfo` to create the file.')
    
    return get_data(json_filepath)

if __name__ == '__main__':
    data_dir = os.path.join(os.path.dirname(__file__), 'data')
    json_filepath = os.path.join(data_dir, 'earth_data.json')

    data = get_data(json_filepath, update=True)
    print(data)





import timeit
from earth_osm.gfk_data import get_region_tuple
from earth_osm.gfk_download import download_pbf, verify_pbf
import signal
import os

def test_download_pbf_update():
    region = get_region_tuple("malta")
    geofabrik_pbf_url = region.urls['pbf']

    data_dir = "earth_data_test"
    fp = os.path.join(data_dir, "pbf", "malta-latest.osm.pbf")
    fp_hash = os.path.join(data_dir, "pbf", "malta-latest.osm.pbf.md5")

    download_pbf(geofabrik_pbf_url, update=True, data_dir=data_dir)

    # check file last modified time
    lm1 = os.path.getmtime(fp)
    print(f"Last modified: {lm1}")

    assert os.path.exists(fp)
    assert verify_pbf(fp, fp_hash)

    download_pbf(geofabrik_pbf_url, update=False, data_dir=data_dir)
    lm2 = os.path.getmtime(fp)
    print(f"Last modified: {lm2}")

    assert lm2 == lm1

    download_pbf(geofabrik_pbf_url, update=True, data_dir=data_dir)
    lm3 = os.path.getmtime(fp)
    print(f"Last modified: {lm3}")

    assert lm3 > lm2


def test_download_corrupted_file():
    data_dir = "earth_data_test"
    region = get_region_tuple("benin")
    geofabrik_pbf_url = region.urls['pbf']
    
    def cancel_download(signal, frame):
        raise KeyboardInterrupt

    signal.signal(signal.SIGALRM, cancel_download)
    signal.alarm(2)

    try:
        download_pbf(geofabrik_pbf_url, update=True, data_dir=data_dir)
    except KeyboardInterrupt:
        print("Download cancelled.")

    
    # check file size of file in human readable format
    fp = os.path.join(data_dir, "pbf", "benin-latest.osm.pbf")
    corrupt_file_size = os.path.getsize(fp)
    print(f"Corrupt File size: {corrupt_file_size / (1024 * 1024)} MB")
    
    download_pbf(geofabrik_pbf_url, update=False, data_dir=data_dir)
    file_size = os.path.getsize(fp)
    print(f"New File size: {file_size / (1024 * 1024)} MB")

    assert file_size > corrupt_file_size


if __name__ == '__main__':
    test_download_pbf_update()
    test_download_corrupted_file()
    


__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Utilities functions for earth_osm

This module contains utilities functions for handling OSM data.

"""

import ast
import logging
import os
import pandas as pd
import geopandas as gpd
from shapely.geometry import LineString, Point, Polygon

from earth_osm import logger as base_logger

logger = logging.getLogger("eo.utils")
logger.setLevel(logging.INFO)

# Notes: Types
# Three main types returned by extraction: Node, Way, Relation
# Node: Point
# Way: LineString or Polygon
# Relation: MultiLineString or MultiPolygon
# So really it is Node, Way, Area, Relation

# Notes: CRS
# geo_crs: EPSG:4326  # general geographic projection, not used for metric measures. "EPSG:4326" is the standard used by OSM and google maps
# distance_crs: EPSG:3857  # projection for distance measurements only. Possible recommended values are "EPSG:3857" (used by OSM and Google Maps)
# area_crs: ESRI:54009  # projection for area measurements only. Possible recommended values are Global Mollweide "ESRI:54009"

def lonlat_lookup(df_way, primary_data):
    """
    Lookup refs and convert to list of longlats
    """
    if "refs" not in df_way.columns:
        logger.warning("refs column not found")
        return []

    def look(ref):
        lonlat_row = list(map(lambda r: tuple(primary_data["Node"][str(r)]["lonlat"]), ref))
        return lonlat_row

    lonlat_list = df_way["refs"].apply(look)

    return lonlat_list

def way_or_area(df_way):
    if "refs" not in df_way.columns:
        raise KeyError("refs column not found")
    
    def check_closed(refs):
        if (refs[0] == refs[-1]) and (len(refs) >= 4):
            return "area"
        elif len(refs) >= 2:
            return "way"
        else:
            logger.debug(f"Way with less than 2 refs: {refs}")
            return None

    type_list = df_way["refs"].apply(check_closed)

    return type_list



def tags_melt(df_exp, nan_threshold=0.75):
    # Find columns with high percentage of NaN values
    high_nan_cols = df_exp.columns[df_exp.isnull().mean() > nan_threshold]

    logger.debug(f"Melting tags from the following columns: {high_nan_cols}")

    df_high_nan = df_exp[high_nan_cols]

    # assert other_tags column does not already exist, if it does,
    assert 'other_tags' not in df_exp.columns, "other_tags column already exists in dataframe"
    df_exp['other_tags'] = df_high_nan.apply(lambda x: x.dropna().to_dict(), axis=1)

    # drop {} in other tags so that its nan
    df_exp['other_tags'] = df_exp['other_tags'].apply(lambda x: x if x != {} else None)


    df_exp.drop(columns=high_nan_cols, inplace=True)
    return df_exp

def columns_melt(df_exp, columns_to_move):
    def concat_melt(row, col):
        # check if value to melt is NaN
        if str(row[col]) == 'nan':
            return None
        # if other_tags column does not exist, no need to concat
        # or if other tags exist, but is empty/none/nan, still no need to concat
        if (
            'other_tags' not in row.keys() or
            row['other_tags'] == {} or
            row['other_tags'] is None or
            str(row['other_tags']) == 'nan'
        ):
            return {col: row[col]}
        else:
            # before concating, check if the column already exists in other_tags
            if col in row['other_tags']:
                logger.warning(f"'{col}' already exists in 'row'.")
            return {**row['other_tags'], col: row[col]}

    # Move specified columns to other_tags
    for col in columns_to_move:
        if col in df_exp.columns:
            df_exp['other_tags'] = df_exp.apply(lambda x: concat_melt(x, col), axis=1)
            df_exp.drop(columns=col, inplace=True)
        else:
            logger.warning(f"Column '{col}' not found in dataframe.")

    return df_exp

def tags_explode(df_melt):
    # check if df_melt has column 'other_tags'
    if not 'other_tags' in df_melt.columns:
        logger.warning("df is not melted, but tags_explode was called")
        return df_melt

    # check if other_tags is empty
    if df_melt['other_tags'].isnull().all():
        logger.debug("nothing to explode, other_tags is empty, returning df with dropped other_tags column")
        # drop other_tags column
        df_melt.drop(columns=['other_tags'], inplace=True)
        return df_melt
    # convert stringified dicts to dict
    df_melt['other_tags'] = df_melt['other_tags'].apply(lambda x: ast.literal_eval(x) if isinstance(x, str) else x)

    # explode other_tags column into multiple columns
    df_exploded = df_melt.join(pd.json_normalize(df_melt['other_tags']))

    # drop the original other_tags column
    df_exploded.drop(columns=['other_tags'], inplace=True)

    return df_exploded


def convert_pd_to_gdf(pd_df):
    # check if pd_df has a lonlat column
    if not 'lonlat' in pd_df.columns:
        raise KeyError("pandas dataframe does not have a lonlat column")
    if not 'Type' in pd_df.columns:
        raise KeyError("pandas dataframe does not have a Type ('node', 'way' or 'area') column")
    # unstringify lonlat column if necessary
    pd_df['lonlat'] = pd_df['lonlat'].apply(lambda x: ast.literal_eval(x) if isinstance(x, str) else x)

    def create_geometry(lonlat_list, geom_type):
        if geom_type == 'node':
            return Point(lonlat_list[0])
        elif geom_type == 'way':
            return LineString(lonlat_list)
        elif geom_type == 'area':
            return Polygon(lonlat_list)
    
    geometry_col = pd_df.apply(lambda row: create_geometry(row['lonlat'], row['Type']), axis=1)
    lonlat_index = pd_df.columns.get_loc('lonlat')
    pd_df.insert(lonlat_index, "geometry", geometry_col)
    gdf = gpd.GeoDataFrame(pd_df, geometry='geometry')
    gdf.drop(columns=['lonlat'], inplace=True)
    pd_df.drop(columns=['geometry'], inplace=True)

    return gdf


def get_list_slug(str_list):
    import hashlib
    str_list.sort()
    if len(str_list) == 1:
        return str_list[0]
    else:
        file_slug = "_".join(str_list)
        if len(file_slug)>15:
            name_string = file_slug[:15] # TODO: could be partial string
            name_code = hashlib.md5(file_slug[15:].encode()).hexdigest()[:8]
            file_slug = name_string + '_' + name_code
        return file_slug

class OutFileWriter:

    def __init__(self, region_list, primary_name, feature_list, data_dir, out_format):
        self.region_list = region_list
        self.primary_name = primary_name
        self.feature_list = feature_list
        self.data_dir = data_dir
        self.out_format = out_format
        logger.info(f'File writer initialized with region_list: {region_list}, primary_name: {primary_name}, feature_list: {feature_list}')

    def __enter__(self):
        # setup file name etc.
        region_slug = get_list_slug(self.region_list) # country code e.g. BJ
        feature_slug = get_list_slug(self.feature_list)
    
        out_dir = os.path.join(self.data_dir, "out")  # Output file directory
        out_slug = os.path.join(out_dir, f"{region_slug}_{feature_slug}")

        self.out_slug = out_slug

        if not os.path.exists(out_dir):
            os.makedirs(
                out_dir, exist_ok=True
            )

        # delete file if it already exists
        if isinstance(self.out_format, str):
            self.out_format = [self.out_format]
        for ext in self.out_format:
            out_path = out_slug + '.' + ext
            if os.path.exists(out_path):
                logger.debug(f"Deleting existing file: {out_path}")
                os.remove(out_path)

        return self
    

    def __call__(self, df_feature):
        if df_feature.empty:
            # exit function
            return

        df_feature.reset_index(drop=True, inplace=True) # avoids weird index

        # Append by concating
        if os.path.exists(self.out_slug + '.csv'):
            # read the existing csv
            df_existing = pd.read_csv(self.out_slug + '.csv')

            # explode other_tags to get back original columns
            df_existing = tags_explode(df_existing)

            # concat the existing df with the new df
            df_feature = pd.concat([df_existing, df_feature], ignore_index=True)

        # melt 95% nan tags (TODO: remove hardcode)
        df_feature = tags_melt(df_feature, 0.95)
        # move refs column to other_tags
        df_feature = columns_melt(df_feature,   ['refs'])

        df_feature.to_csv(self.out_slug + '.csv', index=False)


        # Apend by writing to csv (more efficient, but problem with dynamically changing columns)
        # if os.path.exists(self.out_slug + '.csv'):
        #     # read the existing columns
        #     col_existing = pd.read_csv(self.out_slug + '.csv', nrows=1).columns.to_list()
        #     # check if the columns are the same
        #     col_incoming = df_feature.columns.to_list()
        #     if col_existing != col_incoming:
        #         # add missing columns to the end
        #         missing_cols_in_incoming = [col for col in col_existing if col not in col_incoming]
        #         new_cols_in_incoming = [col for col in col_incoming if col not in col_existing]
        #     TODO: incomplete, should melt new_cols_in_incoming, and add missing cols

        # write the new df to csv
        # df_feature.to_csv(self.out_slug + '.csv', index=False, header= not os.path.exists(self.out_slug + '.csv'), mode="a")


    def __exit__(self, exc_type, exc_value, traceback):
        if os.path.exists(self.out_slug + '.csv'):
            logger.info(f"Output file written to: {self.out_slug}.csv")
            
            if 'geojson' in self.out_format:
                # read the csv file
                df_feature = pd.read_csv(self.out_slug + '.csv')
                # convert to geodataframe
                gdf_feature = convert_pd_to_gdf(df_feature)

                gdf_feature.to_file(
                    self.out_slug + '.geojson', driver="GeoJSON", index=False, mode="w"
                )
                logger.info(f"Output file written to: {self.out_slug}.geojson")


# TODO: deprecate/remove (moved into contextmanager)
def output_creation(df_feature, primary_name, feature_list, region_list, data_dir, out_format):
    """
    Save Dataframe to disk
    Currently supports 
        CSV: Comma Separated Values
        GeoJSON: GeoJSON format (including geometry)

    Args:
        df_feature
    """

    region_slug = get_list_slug(region_list) # country code e.g. BJ
    feature_slug = get_list_slug(feature_list)
    
    out_dir = os.path.join(data_dir, "out")  # Output file directory
    out_slug = os.path.join(out_dir, f"{region_slug}_{feature_slug}")
    

    if not os.path.exists(out_dir):
        os.makedirs(
            out_dir, exist_ok=True
        )  # create raw directory

    df_feature.reset_index(drop=True, inplace=True)

    # Generate Files
    if df_feature.empty:
        logger.warning(f"feature data frame empty for {feature_name}")
        return None

    if "csv" in out_format:
        logger.debug("Writing CSV file")
        df_feature.to_csv(out_slug + '.csv')

    if "geojson" in out_format:
        logger.debug("Writing GeoJSON file")
        gdf_feature = convert_pd_to_gdf(df_feature)
        gdf_feature.to_file(out_slug + '.geojson', driver="GeoJSON")

if __name__ == "__main__":

    from earth_osm.filter import get_filtered_data
    from earth_osm.gfk_data import get_region_tuple
    region = "DE"
    primary_name = "power"
    feature_name = "line"
    mp = True
    update = False
    data_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "earth_data")

    primary_dict, feature_dict = get_filtered_data(get_region_tuple(region), primary_name, feature_name, mp, update, data_dir)

    primary_data = primary_dict['Data']
    feature_data = feature_dict['Data']

    df_node = pd.json_normalize(feature_data["Node"].values())
    df_way = pd.json_normalize(feature_data["Way"].values())

#%%
# sort columns by percentage of nan missing values
# df_feature.isna().mean().sort_values(ascending=True)
# move geometry column to second place
# cols = df_feature.columns.tolist()
# cols.insert(1, cols.pop(cols.index('geometry')))
# df_feature = df_feature.reindex(columns= cols)


#%%
# df_feature.isna().mean()*100
# df_feature.info(memory_usage='deep')
# df_feature['Type'].value_counts(dropna=False)

# drop columns thar are all nan, count them before dropping
# logger.debug(f"Dropping {df_way.isna().all().sum()} columns with all NaN (percentage of columns: {df_way.isna().all().sum()/len(df_way.columns):.2%})")
# df_way.dropna(axis=1, how="all", inplace=True)

# drop columns that have 99% nan values
# logger.debug(f"Dropping {df_way.isna().sum().gt(len(df_way)*0.99).sum()} columns out of {len(df_way.columns)} with more than 99% NaN (percentage of columns: {df_way.isna().sum().gt(len(df_way)*0.99).sum()/len(df_way.columns):.2%})")
# df_way.dropna(axis=1, thresh=len(df_way)*0.01, inplace=True)

# element_set = set(primary_feature_element[primary_name][feature_name])
# print(element_set)
# assert element_set <= set(['node', 'way', 'area']), f"Currenly only supports node, way and area. Got {element_set}"
# if set(['way', 'area']) <= element_set:


__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""
DEPRECATED MODULE (see tagdata.py and taginfo.py)

"""

primary_feature_element = {
    "power": {
        "substation": ("node", "area"), 
        "generator": ("node", "way", "area"),
        "line": ("way"),
        "tower": ("node"),
        "cable": ("way"),
        },
    "building": {
        'allotment_house': ('area',),
        'apartments': ('area',),
        'barn': ('area',),
        'boathouse': ('area',),
        'bungalow': ('area',),
        'bunker': ('area',),
        'cabin': ('area',),
        'carport': ('area',),
        'cathedral': ('area',),
        'chapel': ('area',),
        'church': ('area',),
        'civic': ('area',),
        'college': ('area',),
        'commercial': ('area',),
        'construction': ('area',),
        'cowshed': ('area',),
        'detached': ('area',),
        'dormitory': ('area',),
        'entrance': ('node',),
        'farm': ('area',),
        'farm_auxiliary': ('area',),
        'fire_station': ('area',),
        'garage': ('area',),
        'garages': ('area',),
        'ger': ('area',),
        'grandstand': ('area',),
        'greenhouse': ('area',),
        'hangar': ('area',),
        'hospital': ('area',),
        'hotel': ('area',),
        'house': ('area',),
        'houseboat': ('area',),
        'hut': ('area',),
        'industrial': ('area',),
        'kindergarten': ('area',),
        'manufacture': ('area',),
        'mosque': ('area',),
        'office': ('area',),
        'outbuilding': ('area',),
        'pavilion': ('area',),
        'public': ('area',),
        'residential': ('area',),
        'retail': ('area',),
        'roof': ('area', 'node'),
        'ruins': ('area',),
        'school': ('area',),
        'semidetached_house': ('area',),
        'service': ('area',),
        'shed': ('area',),
        'stable': ('area',),
        'stadium': ('area',),
        'static_caravan': ('area',),
        'sty': ('area',),
        'synagogue': ('area',),
        'temple': ('area',),
        'terrace': ('area',),
        'train_station': ('node', 'area'),
        'transportation': ('area',),
        'university': ('area',),
        'warehouse': ('area',)
        }
    # add more primary features here
}
# %%
# import json
# from pprint import pprint
# import pandas as pd
# TODO: Add taginfo.json to package data, 
# can be used to get the primary keys and values
# TODO: Implement view features simialr to gfk_data
# taginfo_url='https://raw.githubusercontent.com/openstreetmap/id-tagging-schema/main/dist/taginfo.min.json'
# taginfo = '/home/matin/Projects/earth-osm/earth_osm/data/taginfo.min.json'
# with open(taginfo, encoding='utf8') as f:
#     d = json.load(f)
# tag_dict = d['tags']
# pprint(tag_dict)
# tag_df = pd.DataFrame(tag_dict)




import pandas as pd
import logging

from earth_osm.taginfo import get_tag_data
from earth_osm import logger as base_logger

logger = logging.getLogger("eo.tag")
logger.setLevel(logging.INFO)


tag_data = get_tag_data()

def get_feature_list(primary_name):
    return list(tag_data[primary_name]['features'].keys())

def get_primary_list():
    return list(tag_data.keys())

def load_tag_data():
    # load tag_data into a dataframe
    dataframes = []
    json_dict = tag_data
    for primary_key, values in json_dict.items():
        features = values['features']
        for feature_key, feature_values in features.items():
            df = pd.json_normalize(feature_values)
            dataframes.append(df)

    tag_df = pd.concat(dataframes, ignore_index=True)
    return tag_df

#%%
def get_popular_features(primary_name):
    # TODO: check count_all and only return
    # - features that are significant
    # - features that are non-relation
    return get_feature_list(primary_name)
    



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Geofabrik Data Download

This module contains functions to download Geofabrik data.

"""


import hashlib
import logging
import os
import shutil

import requests
import urllib3
from tqdm.auto import tqdm

from earth_osm import logger as base_logger

logger = logging.getLogger("eo.gfk")
logger.setLevel(logging.INFO)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def download_file(url, dir, exists_ok=False):
    """
    Download file from url to dir

    Args:
        url (str): url to download
        dir (str): directory to download to
        exists_ok (bool): Flag to allow skipping download if file exists.

    Returns:
        str: filepath of downloaded file
    """
    filename = os.path.basename(url)
    filepath = os.path.join(dir, filename)
    if os.path.exists(filepath) and exists_ok:
        logger.debug(f'{filepath} already exists')
        return filepath
    logger.info(f"{filename} downloading to {filepath}")
    os.makedirs(os.path.dirname(filepath), exist_ok=True)  #  create download dir
    with requests.get(url, stream=True, verify=False) as r:
        if r.status_code == 200:
            # url properly found, thus execute as expected
            file_size = int(r.headers.get('Content-Length', 0))
            desc = "(Unknown total file size)" if file_size == 0 else ""
            with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc, leave=False) as raw:
                with open(filepath, "wb") as f:
                    shutil.copyfileobj(raw, f)
        else:
            # error status code: file not found
            logger.error(
                f"Error code: {r.status_code}. File {filename} not downloaded from {url}"
            )
            filepath = None
    return filepath

def download_sitemap(geom, pkg_data_dir):
    geofabrik_geo = "https://download.geofabrik.de/index-v1.json"
    geofabrik_nogeo = "https://download.geofabrik.de/index-v1-nogeom.json"
    geofabrik_sitemap_url = geofabrik_geo if geom else geofabrik_nogeo

    sitemap_file = download_file(geofabrik_sitemap_url, pkg_data_dir, exists_ok=True)

    return sitemap_file


def download_pbf(url, update, data_dir):

    pbf_dir = os.path.join(data_dir, "pbf")
    pbf_fn = os.path.basename(url)
    pbf_fp = os.path.join(pbf_dir, pbf_fn)

    # download file
    down_pbf_fp = download_file(url, pbf_dir, exists_ok=not update)
    down_md5_fp = download_file(url + ".md5", pbf_dir, exists_ok=not update)
    
    assert down_pbf_fp == pbf_fp

    if not verify_pbf(down_pbf_fp, down_md5_fp):
        logger.info(f"PBF Md5 mismatch, retrying download for {pbf_fn}")
        down_pbf_fp = download_file(url, pbf_dir)
        down_md5_fp = download_file(url + ".md5", pbf_dir, exists_ok=not update)
        if not verify_pbf(down_pbf_fp, down_md5_fp):
            raise ValueError(f"File verification failed after retry for {pbf_fn}")

    return pbf_fp


def calculate_md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def verify_pbf(pbf_inputfile, pbf_md5file):
    # Calculate local MD5
    local_md5 = calculate_md5(pbf_inputfile)

    # Read remote MD5
    with open(pbf_md5file, 'r') as f:
        remote_md5 = f.read().split()[0]

    return local_md5 == remote_md5



import os
from earth_osm.eo import save_osm_data


primary_name = "power"
update = False
mp = True
data_dir = os.path.join(os.getcwd(), "earth_data_test")
out_dir = os.path.join(os.getcwd(), "earth_data_test")


def test_no_aggregate(shared_data_dir):
    save_osm_data(
        region_list=["nigeria", "benin"],
        primary_name=primary_name,
        feature_list=["substation", "generator", "line", "cable"],
        update=update,
        mp=mp,
        data_dir=shared_data_dir,
        out_dir=out_dir,
        out_format=["csv", "geojson"],
        out_aggregate=False,
    )
    assert True

# same as pypsa-tutorial
def test_aggregate(shared_data_dir):
    save_osm_data(
        region_list=["benin", "nigeria"],
        primary_name=primary_name,
        feature_list=["substation", "line", "cable", "generator"],
        update=update,
        mp=mp,
        data_dir=data_dir,
        out_dir=out_dir,
        out_format=["csv", "geojson"],
        out_aggregate=True,
    )
    assert True

def test_small_count(shared_data_dir):
    save_osm_data(
        region_list=["malta"],
        primary_name="power",
        feature_list=["plant"],
        update=update,
        mp=mp,
        data_dir=data_dir,
        out_dir=out_dir,
        out_format=["csv", "geojson"],
        out_aggregate=False,
    )
    assert True

# test low resource feature (cable)
# test high resource feature (substation)

# test aggregation by region (benin, germany)
# test aggregation by feature (tower, line)
# test no aggregation
 
 


__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""
This is the principal module of the earth_osm project.
"""

import logging
import os

import pandas as pd

# suppress pandas warning about fragmented dataframes
# TODO: do not suppress warnings
import warnings
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

from earth_osm.tagdata import get_feature_list
from earth_osm.filter import get_filtered_data
from earth_osm.gfk_data import get_region_tuple, view_regions
from earth_osm.utils import OutFileWriter, lonlat_lookup, way_or_area
from earth_osm import logger as base_logger

logger = logging.getLogger("eo.eo")
logger.setLevel(logging.INFO)

def process_region(region, primary_name, feature_name, mp, update, data_dir):
    """
    Process Country

    Args:
        region: Region object
        primary_name: Primary Feature Name
        feature_name: Feature Name
        mp: Multiprocessing object
        update: Update flag

    Returns:
        None
    """
    primary_dict, feature_dict = get_filtered_data(region, primary_name, feature_name, mp, update, data_dir)

    primary_data = primary_dict['Data']
    feature_data = feature_dict['Data']

    df_node = pd.json_normalize(feature_data["Node"].values())
    df_way = pd.json_normalize(feature_data["Way"].values())

    if df_way.empty:
        logger.debug(f"df_way is empty for {region.short}, {primary_name}, {feature_name}")
        # for df_way, check if way or area
    else:
        type_col = way_or_area(df_way)
        df_way.insert(1, "Type", type_col)
        logger.debug(df_way['Type'].value_counts(dropna=False))

        # Drop rows with None in Type
        logger.debug(f"Dropping {df_way['Type'].isna().sum()} rows with None in Type")
        df_way.dropna(subset=["Type"], inplace=True)

        # convert refs to lonlat
        lonlat_column = lonlat_lookup(df_way, primary_data)
        df_way.insert(1, "lonlat", lonlat_column)

    # check if df_node is empty
    if df_node.empty:
        logger.debug(f"df_node is empty for {region.short}, {primary_name}, {feature_name}")
    else:
        # df node has lonlat as [lon, lat] it should be [(lon, lat)]
        df_node["lonlat"] = df_node["lonlat"].apply(lambda x: [tuple(x)])
        
        # set type to node
        df_node["Type"] = "node"
    
    # concat ways and nodes
    df_feature = pd.concat([df_way, df_node], ignore_index=True)

    # remove columns that are all nan
    df_feature.dropna(axis=1, how="all", inplace=True)

    if df_feature.empty:
        logger.debug(f"df_feature is empty for {region.short}, {primary_name}, {feature_name}")
    else:
        df_feature.insert(3, 'Region', region.short)

    # dev logger warning
    if 'other_tags' in df_feature.columns:
        logger.warning(f"other_tags in extracted data from osm, change of other_tags to eo_tags is necessary, please open issue on github")
        
    return df_feature

def get_osm_data(
        region_str,
        primary_name,
        feature_name,
        data_dir=None,
        cached = True):
    
    region_tuple = get_region_tuple(region_str)
    mp = True
    update = not cached

    data_dir=os.path.join(os.getcwd(), 'earth_data') if data_dir is None else data_dir
    
    df = process_region(
        region_tuple,
        primary_name,
        feature_name,
        mp,
        update,
        data_dir,
    )

    return df
    


# TODO: Plan
# Use an intermediary super efficient file format such as parquet
# save a region,feauture pair in temp files
# implement planetary file from osm (https://planet.openstreetmap.org/)
# read keys, values, tags and their frequencies from taginfo api (https://taginfo.openstreetmap.org/taginfo/apidoc)
# use **kwargs and allow for dropping of refs column
# implement post processing functions: i) create_geojson 2) filter_by_bbox


def save_osm_data(
    region_list,
    primary_name,
    feature_list=None,
    update=False,
    mp=True, # TODO: remove mp arg
    data_dir=os.path.join(os.getcwd(), 'earth_data'),
    out_dir=os.path.join(os.getcwd(), 'earth_data'),
    out_format="csv", # TODO: rename out_format -> format
    out_aggregate=True, # TODO: rename out_aggregate -> aggregate
):
    """
    Get OSM Data for a list of regions and features
    args:
        region_list: list of regions to get data for
        primary_name: primary feature to get data for
        feature_list: list of features to get data for
        update: update data
        mp: use multiprocessing
    returns:
        dict of dataframes
    """
    region_tuple_list = [get_region_tuple(r) for r in region_list]
    region_short_list = [r.short for r in region_tuple_list]

    if feature_list is None:
        feature_list = get_feature_list(primary_name)
    elif feature_list == ['ALL']:
        # Account for wild card
        feature_list = [f'ALL_{primary_name}']

    if out_aggregate == 'region' or out_aggregate is True:
        # for each feature, aggregate all regions
        for feature_name in feature_list:
            with OutFileWriter(region_short_list, primary_name, [feature_name], out_dir, out_format) as writer:
                for region in region_tuple_list:
                    df_feature = process_region(region, primary_name, feature_name, mp, update, data_dir)
                    writer(df_feature)
                    # output_creation(df_feature, primary_name, [feature_name], region_short_list, data_dir, out_format)

    elif out_aggregate == 'feature':
        # for each region, aggreagate all features
        for region in region_tuple_list:
            with OutFileWriter([region.short], primary_name, feature_list, out_dir, out_format) as writer:
                for feature_name in feature_list:
                    df_feature = process_region(region, primary_name, feature_name, mp, update, data_dir)
                    writer(df_feature)
    
    elif out_aggregate is False:
        # no aggregation, one file per region per feature
        for region in region_tuple_list:
                for feature_name in feature_list:
                    df_feature = process_region(region, primary_name, feature_name, mp, update, data_dir)
                    with OutFileWriter([region.short], primary_name, [feature_name], out_dir, out_format) as writer:
                        writer(df_feature)


    # combinations = ((region, feature_name) for region in region_tuple_list for feature_name in feature_list)

    # processed_data = map(lambda combo: process_region(combo[0], primary_name, combo[1], mp, update, data_dir), combinations)

    # for i, combo in enumerate(combinations):
    #     output_creation(processed_data[i], primary_name, combo[1], [combo[0]], data_dir, out_format, out_aggregate)

